@@ -1,7 +1,7 @@
 /*
  * fsarchiver: Filesystem Archiver
  *
- * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
+ * Copyright (C) 2008-2010 Francois Dupoux.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
@@ -15,28 +15,29 @@
  * Homepage: http://www.fsarchiver.org
  */
 
-#include "fsarchiver.h"
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
 
 #include <stdarg.h>
-#include <stdio.h>
 #include <errno.h>
 #include <string.h>
 #include <unistd.h>
-#include <linux/limits.h>
 #include <dirent.h>
 #include <sys/time.h>
-#include <sys/types.h>
 #include <sys/wait.h>
-#include <time.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <stdlib.h>
 #include <execinfo.h>
 #include <wordexp.h>
+#include <fnmatch.h>
+#include <time.h>
 
+#include "fsarchiver.h"
 #include "syncthread.h"
+#include "strlist.h"
 #include "common.h"
+#include "error.h"
 
 int stream_readline(FILE *f, char *buf, int buflen)
 {
@@ -59,7 +60,7 @@ void concatenate_paths(char *buffer, int maxbufsize, char *p1, char *p2)
     *(buffer++)='/';
     while (*p2=='/')
         p2++;
-    strlcat(buffer, p2, maxbufsize);
+    strlcatf(buffer, maxbufsize, "%s", p2);
 }
 
 char *format_size(u64 size, char *text, int max, char units)
@@ -130,6 +131,18 @@ int extract_dirpath(char *filepath, char *dirbuf, int dirbufsize)
     return 0;
 }
 
+int extract_basename(char *filepath, char *basenamebuf, int basenamebufsize)
+{
+    int i;
+
+    for (i=0; filepath[i]; i++);
+    while ((i>0) && (filepath[i]!='/') && (filepath[i-1]!='/'))
+        i--;
+    snprintf(basenamebuf, basenamebufsize, "%s", &filepath[i]);
+
+    return 0;
+}
+
 char *get_objtype_name(int objtype)
 {
     switch (objtype)
@@ -237,18 +250,46 @@ int regfile_exists(char *filepath)
     return -1; // don't know
 }
 
-int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize, char *stderrbuf, int stderrsize, char *format, ...)
+int getpathtoprog(char *buffer, int bufsize, char *prog)
 {
+    char pathtest[PATH_MAX];
+    char delims[]=":\t\n";
+    struct stat bufstat;
+    char pathenv[4096];
+    char *saveptr=0;
+    char *result;
+    char *vp;
+    int i;
+    
+    memset(buffer, 0, bufsize);
+    if ((vp=getenv("PATH")) == NULL)
+        return -1;
+    snprintf(pathenv, sizeof(pathenv), "%s", vp);
+    result=strtok_r(pathenv, delims, &saveptr);
+    for(i=0; result != NULL; i++)
+    {
+        snprintf(pathtest, sizeof(pathtest), "%s/%s", result, prog);
+        if (stat(pathtest, &bufstat)==0 && access(pathtest, X_OK)==0)
+        {
+            snprintf(buffer, bufsize, "%s", pathtest);
+            return 0;
+        }
+        result = strtok_r(NULL, delims, &saveptr);
+    }
+    return -1;
+}
+
+int exec_command(char *command, int cmdbufsize, int *exitst, char *stdoutbuf, int stdoutsize, char *stderrbuf, int stderrsize, char *format, ...)
+{
+    char pathtoprog[PATH_MAX]; // full path to the program to run
     const int max_argv=128; // maximum arguments to a command
     char *argv[max_argv]; // pointer to arguments processed by wordexp()
     int outpos=0; // how many bytes have already been stored in the buffer for stdout
     int errpos=0; // how many bytes have already been stored in the buffer for stderr
-    char temp[2048];
     int pfildes1[2];
     int pfildes2[2];
     wordexp_t p;
     int status;
-    int exitst;
     int mystdout;
     int mystderr;
     va_list ap;
@@ -260,19 +301,21 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
     int i;
     
     // init
+    memset(pathtoprog, 0, sizeof(pathtoprog));
     for (i=0; i < max_argv; argv[i++]=NULL);
+    if (exitst)
+        *exitst=-1;
     
     // format the string
-    memset(stdoutbuf, 0, stdoutsize);
-    memset(stderrbuf, 0, stderrsize);
+    if (stdoutbuf && stdoutsize)
+        memset(stdoutbuf, 0, stdoutsize);
+    if (stderrbuf && stderrsize)
+        memset(stderrbuf, 0, stderrsize);
     memset(command, 0, cmdbufsize);
     va_start(ap, format);
     vsnprintf(command, cmdbufsize, format, ap);
     va_end(ap);
     
-    // parse the command
-    //snprintf(temp, sizeof(temp), "%s", command);
-    
     // do shell expansion to parse the quotes for args with spaces
     wordexp(command, &p, 0); // will require wordfree to free the memory
     for(i=0; (i < p.we_wordc) && (i<max_argv); i++)
@@ -280,15 +323,13 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
         msgprintf(MSG_DEBUG1, "argv[%d]=[%s]\n", i, argv[i]);
     }
     
-    // make sure the program exists
-    snprintf(temp, sizeof(temp), "which %s 1>/dev/null 2>/dev/null", argv[0]);
-    res=system(temp);
-    msgprintf(MSG_DEBUG1, "system(%s)=%d\n", temp, res);
-    if (res!=0)
-    {   errprintf("program [%s] cannot be executed\n", argv[0]);
+    // get the full path to the program
+    if (getpathtoprog(pathtoprog, sizeof(pathtoprog), argv[0])!=0)
+    {   errprintf("program [%s] no found in PATH or bad permissions on that program\n", argv[0]);
         wordfree(&p);
         return -1;
     }
+    msgprintf(MSG_VERB2, "getpathtoprog(%s)=[%s]\n", argv[0], pathtoprog);
     
     // execute the command
     if ((pipe(pfildes1)==-1) || (pipe(pfildes2)==-1))
@@ -300,7 +341,7 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
     msgprintf(MSG_VERB1, "executing [%s]...\n", command);
     
     if ((pid = fork()) == -1)
-    {   errprintf("fork() failed\n");
+    {   sysprintf("fork() failed\n");
         wordfree(&p);
         return -1;
     }
@@ -312,8 +353,8 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
         close(pfildes1[1]); // close excess fildes
         dup2(pfildes2[1],2); // make 1 same as write-to end of pipe
         close(pfildes2[1]); // close excess fildes
-        execvp(argv[0], argv);
-        errprintf("execvp(%s) failed\n", argv[0]); // still around? exec failed
+        execvp(pathtoprog, argv);
+        errprintf("execvp(%s) failed\n", pathtoprog); // still around? exec failed
         wordfree(&p);
         exit(EXIT_FAILURE);
     }
@@ -333,14 +374,14 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
         do
         {   // read data stored in the stdout pipe (read to prevents the sub-process command from blocking on write)
             do
-            {    res=read(mystdout, &c, 1);
+            {   res=read(mystdout, &c, 1);
                 if ((stdoutbuf!=NULL) && (res>0) && (outpos+1 < stdoutsize))
                     stdoutbuf[outpos++]=c;
             } while (res>0);
             
             // read data stored in the stderr pipe (read to prevents the sub-process command from blocking on write)
             do
-            {    res=read(mystderr, &c, 1);
+            {   res=read(mystderr, &c, 1);
                 if ((stderrbuf!=NULL) && (res>0) && (errpos+1 < stderrsize))
                     stderrbuf[errpos++]=c;
             } while (res>0);
@@ -354,9 +395,9 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
         if ((stderrbuf!=NULL) && (errpos+1 < stderrsize))
             res=read(mystderr, stderrbuf+errpos, stderrsize-errpos-1);
         
-        exitst=WEXITSTATUS(status);
-        
-        msgprintf(MSG_VERB1, "command [%s] returned %d\n", command, exitst);
+        msgprintf(MSG_VERB1, "command [%s] returned %d\n", command, WEXITSTATUS(status));
+        if (exitst)
+            *exitst=WEXITSTATUS(status);
         
         if ((stdoutbuf!=NULL) && (outpos>0))
             msgprintf(MSG_DEBUG1, "\n----stdout----\n%s\n----stdout----\n\n", stdoutbuf);
@@ -364,14 +405,14 @@ int exec_command(char *command, int cmdbufsize, char *stdoutbuf, int stdoutsize,
             msgprintf(MSG_DEBUG1, "\n----stderr----\n%s\n----stderr----\n\n", stderrbuf);
         
         wordfree(&p);
-        return exitst;
+        return 0;
     }
 }
 
 int generate_random_tmpdir(char *buffer, int bufsize, int n)
 {
-        struct tm tbreak;
-        time_t abstime;
+    struct tm tbreak;
+    time_t abstime;
     struct timeval tv;
     
     gettimeofday(&tv, NULL);
@@ -437,23 +478,12 @@ char *format_time(char *buffer, int bufsize, u64 t)
     t2=t;
     
     localtime_r(&t2, &timeres);
-    snprintf(buffer, bufsize, "%.4d%.2d%.2d-%.2d:%.2d:%.2d",
+    snprintf(buffer, bufsize, "%.4d-%.2d-%.2d_%.2d-%.2d-%.2d",
         timeres.tm_year+1900, timeres.tm_mon+1, timeres.tm_mday,
         timeres.tm_hour, timeres.tm_min, timeres.tm_sec);
     return buffer;
 }
 
-// replacement for strncat() where we can't specify the destbuf size
-char *strlcat(char *dest, const char *src, int n)
-{
-    int i, pos;
-    
-    for (i=strlen(dest), pos=0; (i < n-1) && (src[pos] != 0) ; i++, pos++)
-        dest[i] = src[pos];
-    dest[i]=0;
-    return dest;
-}
-
 // add a formatted string at the end of a buffer that already contains a string
 char *strlcatf(char *dest, int destbufsize, char *format, ...)
 {
@@ -520,20 +550,62 @@ int format_stacktrace(char *buffer, int bufsize)
 {
     const int stack_depth=20;
     void *temp[stack_depth];
-        char **strings;
-        int nptrs;
-        int i;
+    char **strings;
+    int nptrs;
+    int i;
     
-        // format the backtrace (advanced error info)
-        memset(buffer, 0, bufsize);
-        nptrs=backtrace(temp, stack_depth);
-        strings=backtrace_symbols(temp, nptrs);
-        if (strings!=NULL)
-        {
-                for (i = 0; i < nptrs; i++)
-                        strlcatf(buffer, bufsize, "%s\n", strings[i]);
-                free(strings);
-        }
+    // format the backtrace (advanced error info)
+    memset(buffer, 0, bufsize);
+    nptrs=backtrace(temp, stack_depth);
+    strings=backtrace_symbols(temp, nptrs);
+    if (strings!=NULL)
+    {
+        for (i = 0; i < nptrs; i++)
+            strlcatf(buffer, bufsize, "%s\n", strings[i]);
+        free(strings);
+    }
     
-        return 0;
+    return 0;
 }
+
+int exclude_check(cstrlist *patlist, char *string)
+{
+    char pattern[1024];
+    int count;
+    int i;
+    
+    count=strlist_count(patlist);
+    for (i=0; i < count; i++)
+    {
+        strlist_getitem(patlist, i, pattern, sizeof(pattern));
+        if (fnmatch(pattern, string, 0)==0)
+            return true;
+    }
+    return false;
+}
+
+int get_path_to_volume(char *newvolbuf, int bufsize, char *basepath, long curvol)
+{
+    char prefix[PATH_MAX];
+    int pathlen;
+    
+    if ((pathlen=strlen(basepath))<4) // all archives terminates with ".fsa"
+    {   errprintf("archive has an invalid basepath: [%s]\n", basepath);
+        return -1;
+    }
+    
+    if (curvol==0) // first volume
+    {
+        if (realpath(basepath, newvolbuf)!=newvolbuf)
+            snprintf(newvolbuf, bufsize, "%s", basepath);
+    }
+    else // not the first volume
+    {
+        memset(prefix, 0, sizeof(prefix));
+        memcpy(prefix, basepath, pathlen-2);
+        snprintf(newvolbuf, bufsize, "%s%.2ld", prefix, (long)curvol);
+    }
+    
+    return 0;
+}
+
